/// Environment variable name aliases.
/// Format: ("<Current Env Name>", "<Alternate name to check>")
/// These aliases support backward compatibility with old environment variable names
/// from before the config refactor (when constants were in data/src/constants.rs and cas_client/src/constants.rs).
pub const ENVIRONMENT_NAME_ALIASES: &[(&str, &str)] = &[
    // Older environment variable names that we should check for backward compatibility.
    ("HF_XET_CLIENT_NUM_RANGE_IN_SEGMENT_BASE", "HF_XET_NUM_RANGE_IN_SEGMENT_BASE"),
    ("HF_XET_CLIENT_NUM_RANGE_IN_SEGMENT_DELTA", "HF_XET_NUM_RANGE_IN_SEGMENT_DELTA"),
    ("HF_XET_CLIENT_NUM_RANGE_IN_SEGMENT_MAX", "HF_XET_NUM_RANGE_IN_SEGMENT_MAX"),
    ("HF_XET_CLIENT_NUM_CONCURRENT_RANGE_GETS", "HF_XET_NUM_CONCURRENT_RANGE_GETS"),
    ("HF_XET_CLIENT_UPLOAD_REPORTING_BLOCK_SIZE", "HF_XET_UPLOAD_REPORTING_BLOCK_SIZE"),
    ("HF_XET_CLIENT_RECONSTRUCT_WRITE_SEQUENTIALLY", "HF_XET_RECONSTRUCT_WRITE_SEQUENTIALLY"),
    (
        "HF_XET_DATA_MIN_SPACING_BETWEEN_GLOBAL_DEDUP_QUERIES",
        "HF_XET_MIN_SPACING_BETWEEN_GLOBAL_DEDUP_QUERIES",
    ),
    ("HF_XET_DATA_LOCAL_CAS_SCHEME", "HF_XET_LOCAL_CAS_SCHEME"),
    ("HF_XET_DATA_CURRENT_VERSION", "HF_XET_CURRENT_VERSION"),
    ("HF_XET_CLIENT_FIXED_CONCURRENCY_MAX_UPLOADS", "HF_XET_MAX_CONCURRENT_UPLOADS"),
    ("HF_XET_CLIENT_FIXED_CONCURRENCY_MAX_DOWNLOADS", "HF_XET_MAX_CONCURRENT_DOWNLOADS"),
    ("HF_XET_DATA_MAX_CONCURRENT_FILE_INGESTION", "HF_XET_MAX_CONCURRENT_FILE_INGESTION"),
    ("HF_XET_DATA_INGESTION_BLOCK_SIZE", "HF_XET_INGESTION_BLOCK_SIZE"),
    ("HF_XET_DATA_PROGRESS_UPDATE_INTERVAL", "HF_XET_PROGRESS_UPDATE_INTERVAL"),
    ("HF_XET_DATA_PROGRESS_UPDATE_SPEED_SAMPLING_WINDOW", "HF_XET_PROGRESS_UPDATE_SPEED_SAMPLING_WINDOW"),
    ("HF_XET_DATA_SESSION_XORB_METADATA_FLUSH_INTERVAL", "HF_XET_SESSION_XORB_METADATA_FLUSH_INTERVAL"),
    ("HF_XET_DATA_SESSION_XORB_METADATA_FLUSH_MAX_COUNT", "HF_XET_SESSION_XORB_METADATA_FLUSH_MAX_COUNT"),
    ("HF_XET_MDB_SHARD_CACHE_SIZE_LIMIT", "HF_XET_SHARD_CACHE_SIZE_LIMIT"),
    ("HF_XET_MDB_SHARD_CHUNK_INDEX_TABLE_MAX_SIZE", "HF_XET_CHUNK_INDEX_TABLE_MAX_SIZE"),
];
